#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module implements data loading utilities for CKIPNLP.
"""

# pylint: disable=missing-docstring

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

import os as _os
import shutil as _shutil

from abc import (
    ABCMeta as _ABCMeta,
    abstractmethod as _abstractmethod,
)

from appdirs import (
    AppDirs as _AppDirs,
)

from ckipnlp.util.logger import (
    get_logger as _get_logger
)

_ROOTDIRS = _AppDirs(appname='ckipnlp', appauthor='ckip')

################################################################################################################################

class _DataBase(metaclass=_ABCMeta):

    name = NotImplemented
    fullname = NotImplemented
    env = NotImplemented
    extra_dirs = ()

    @classmethod
    def env_data_dir(cls):
        return _os.getenv(cls.env)  # pylint: disable=invalid-envvar-value

    @classmethod
    def user_data_dir(cls):
        return _os.path.join(_ROOTDIRS.user_data_dir, cls.name)

    @classmethod
    def site_data_dir(cls):
        return _os.path.join(_ROOTDIRS.user_data_dir, cls.name)

    @classmethod
    def get_data(cls):
        for data_dir in (
            cls.env_data_dir(),
            cls.user_data_dir(),
            cls.site_data_dir(),
            *cls.extra_dirs,
        ):
            if data_dir and _os.path.isdir(data_dir):
                break
        else:
            _get_logger().warning(f'No existing data for {cls.fullname}. Download data from remote ...')
            data_dir = cls.download_data()
        return data_dir

    @classmethod
    def install_data(cls, src_dir, *, copy=False):
        data_dir = cls.user_data_dir()

        if _os.path.isdir(data_dir):
            _get_logger().warning(f'{data_dir} already exists!')

        else:
            _os.makedirs(_os.path.dirname(data_dir), exist_ok=True)
            if not copy:
                _get_logger().debug(f'Linking CkipTagger data from {src_dir} to {data_dir} ...')
                _os.symlink(src_dir, data_dir)
                _get_logger().debug(f'Linking CkipTagger data from {src_dir} to {data_dir} ... done')
            else:
                _get_logger().debug(f'Copying CkipTagger data from {src_dir} to {data_dir} ...')
                _shutil.copytree(src_dir, data_dir)
                _get_logger().debug(f'Copying CkipTagger data from {src_dir} to {data_dir} ... done')

        return data_dir

    @classmethod
    def download_data(cls):
        data_dir = cls.user_data_dir()
        _get_logger().debug(f'Downloading CkipTagger data to {data_dir} ...')
        cls._download_data(data_dir)
        _get_logger().debug(f'Downloading CkipTagger data to {data_dir} ... done')
        return data_dir

    @classmethod
    @_abstractmethod
    def _download_data(cls, data_dir):
        return NotImplemented


################################################################################################################################

class TaggerData(_DataBase):

    name = 'tagger'
    fullname = 'CkipTagger'
    env = 'CKIPTAGGER_DATA'
    extra_dirs = ('./data',)

    @classmethod
    def user_data_dir(cls):
        return _os.path.join(super().user_data_dir(), 'data')

    @classmethod
    def site_data_dir(cls):
        return _os.path.join(super().site_data_dir(), 'data')

    @classmethod
    def _download_data(cls, data_dir):
        data_dir = super().user_data_dir()
        from ckiptagger.data_utils import download_data_url
        _os.makedirs(data_dir, exist_ok=True)
        download_data_url(data_dir)

# pylint: disable=invalid-name
get_tagger_data = TaggerData.get_data            #: Get CkipTagger data directory.
install_tagger_data = TaggerData.install_data    #: Link/Copy CkipTagger data directory.
download_tagger_data = TaggerData.download_data  #: Download CkipTagger data directory.



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides containers for word-segmented sentences with part-of-speech-tags.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from abc import (
    ABCMeta as _ABCMeta,
    abstractmethod as _abstractmethod,
)

from typing import (
    NamedTuple as _NamedTuple,
)

from ..base import (
    BaseTuple as _BaseTuple,
)

from ..seg import (
    SegSentence as _SegSentence,
    SegParagraph as _SegParagraph,
)

################################################################################################################################

def _token_from_text(data):
    """str -> Tuple[str, str]"""
    return data.strip().strip(')').rsplit('(', 1)

def _token_to_text(data):
    """Tuple[str, str] -> str"""
    return '{}({})'.format(*data)

def _sentence_from_text(data):
    """str -> Tuple[Iterable[str], Iterable[str]]"""
    return zip(*map(_token_from_text, data.split('\u3000')))

def _sentence_to_text(data):
    """Tuple[Iterable[str], Iterable[str]] -> str"""
    return '\u3000'.join(map(_token_to_text, zip(*data)))

def _paragraph_from_text(data):
    """Iterable[str] -> Tuple[Iterable[Iterable[str]], Iterable[Iterable[str]]]"""
    return zip(*map(_sentence_from_text, data))

def _paragraph_to_text(data):
    """Tuple[Iterable[Iterable[str]], Iterable[Iterable[str]]] -> Iterable[str]"""
    return map(_sentence_to_text, zip(*data))

################################################################################################################################

class _WsPosToken(_NamedTuple):
    word: str = None
    pos: str = None

class WsPosToken(_BaseTuple, _WsPosToken):
    """A word with POS-tag.

    Attributes
    ----------
        word : str
            the word.
        pos : str
            the POS-tag.

    Note
    ----
        This class is an subclass of *tuple*. To change the attribute, please create a new instance instead.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`from_text` and :meth:`to_text`.

            .. code-block:: python

                '中文字(Na)'  # word / POS-tag

        List format
            Used for :meth:`from_list` and :meth:`to_list`.

            .. code-block:: python

                [
                    '中文字', # word
                    'Na',    # POS-tag
                ]

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.

            .. code-block:: python

                {
                    'word': '中文字', # word
                    'pos': 'Na',     # POS-tag
                }
    """

    def __str__(self):
        return self.to_text()

    ########################################################################################################################

    @classmethod
    def from_text(cls, data):
        """Construct an instance from text format.

        Parameters
        ----------
            data : str
                text such as ``'中文字(Na)'``.

        .. note::
            - ``'中文字(Na)'`` -> word = ``'中文字'``, pos = ``'Na'``
            - ``'中文字'``     -> word = ``'中文字'``, pos = ``None``
        """
        return cls(*_token_from_text(data))

    def to_text(self):
        return _token_to_text(self)

################################################################################################################################

class WsPosSentence(metaclass=_ABCMeta):
    """A helper class for data conversion of word-segmented and part-of-speech sentences."""

    @_abstractmethod
    def __init__(self):
        raise NotImplementedError  # pragma: no cover

    ########################################################################################################################

    @classmethod
    def from_text(cls, data):
        """Convert text format to word-segmented and part-of-speech sentences.

        Parameters
        ----------
            data : str
                text such as ``'中文字(Na)\\u3000耶(T)'``.

        Returns
        -------
            :class:`~ckipnlp.container.seg.SegSentence`
                the word sentence
            :class:`~ckipnlp.container.seg.SegSentence`
                the POS-tag sentence.
        """
        return tuple(map(_SegSentence.from_list, _sentence_from_text(data)))

    @staticmethod
    def to_text(word, pos):
        """Convert text format to word-segmented and part-of-speech sentences.

        Parameters
        ----------
            word : :class:`~ckipnlp.container.seg.SegSentence`
                the word sentence
            pos  : :class:`~ckipnlp.container.seg.SegSentence`
                the POS-tag sentence.

        Returns
        -------
            str
                text such as ``'中文字(Na)\\u3000耶(T)'``.
        """
        return _sentence_to_text((word, pos,))

################################################################################################################################

class WsPosParagraph(metaclass=_ABCMeta):
    """A helper class for data conversion of word-segmented and part-of-speech sentence lists."""

    @_abstractmethod
    def __init__(self):
        raise NotImplementedError  # pragma: no cover

    ########################################################################################################################

    @classmethod
    def from_text(cls, data):
        """Convert text format to word-segmented and part-of-speech sentence lists.

        Parameters
        ----------
            data : Sequence[str]
                list of sentences such as ``'中文字(Na)\\u3000耶(T)'``.

        Returns
        -------
            :class:`~.seg.SegParagraph`:
                the word sentence list
            :class:`~.seg.SegParagraph`:
                the POS-tag sentence list.
        """
        return tuple(map(_SegParagraph.from_list, _paragraph_from_text(data)))

    @staticmethod
    def to_text(word, pos):
        """Convert text format to word-segmented and part-of-speech sentence lists.

        Parameters
        ----------
            word : :class:`~.seg.SegParagraph`
                the word sentence list
            pos  : :class:`~.seg.SegParagraph`
                the POS-tag sentence list.

        Returns
        -------
            List[str]
                list of sentences such as ``'中文字(Na)\\u3000耶(T)'``.
        """
        return list(_paragraph_to_text((word, pos,)))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides containers for NER sentences.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from typing import (
    NamedTuple as _NamedTuple,
    Tuple as _Tuple,
)

from .base import (
    BaseTuple as _BaseTuple,
    BaseList as _BaseList,
    BaseSentence as _BaseSentence,
)

################################################################################################################################

class _NerToken(_NamedTuple):
    word: str
    ner: str
    idx: _Tuple[int, int]

class NerToken(_BaseTuple, _NerToken):
    """A named-entity recognition token.

    Attributes
    ----------
        word : str
            the token word.
        ner : str
            the NER-tag.
        idx : Tuple[int, int]
            the starting / ending index.

    Note
    ----
        This class is an subclass of :class:`tuple`. To change the attribute, please create a new instance instead.

    .. admonition:: Data Structure Examples

        Text format
            Not implemented

        List format
            Used for :meth:`from_list` and :meth:`to_list`.

            .. code-block:: python

                [
                    '中文字'     # token word
                    'LANGUAGE', # NER-tag
                    (0, 3),     # starting / ending index.
                ]

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.

            .. code-block:: python

                {
                    'word': '中文字',   # token word
                    'ner': 'LANGUAGE', # NER-tag
                    'idx': (0, 3),     # starting / ending index.
                }

        CkipTagger format
            Used for :meth:`from_tagger` and :meth:`to_tagger`.

            .. code-block:: python

                (
                    0,          # starting index
                    3,          # ending index
                    'LANGUAGE', # NER-tag
                    '中文字',    # token word
                )
    """

    def __new__(cls, word, ner, idx, **kwargs):  # pylint: disable=signature-differs
        return super().__new__(cls, word, ner, tuple(idx), **kwargs)

    to_text = NotImplemented
    from_text = NotImplemented

    ########################################################################################################################

    @classmethod
    def from_tagger(cls, data):
        """Construct an instance from CkipTagger format."""
        idx0, idx1, ner, word = data
        return cls(word=word, ner=ner, idx=(idx0, idx1,))  # pylint: disable=no-value-for-parameter

    def to_tagger(self):
        """Transform to CkipTagger format."""
        return (*self.idx, self.ner, self.word)

################################################################################################################################

class NerSentence(_BaseSentence):
    """A named-entity recognition sentence.

    .. admonition:: Data Structure Examples

        Text format
            Not implemented

        List format
            Used for :meth:`from_list` and :meth:`to_list`.

            .. code-block:: python

                [
                    [ '美國', 'GPE', (0, 2), ],   # name-entity 1
                    [ '參議院', 'ORG', (3, 5), ], # name-entity 2
                ]

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.

            .. code-block:: python

                [
                    { 'word': '美國', 'ner': 'GPE', 'idx': (0, 2), },   # name-entity 1
                    { 'word': '參議院', 'ner': 'ORG', 'idx': (3, 5), }, # name-entity 2
                ]

        CkipTagger format
            Used for :meth:`from_tagger` and :meth:`to_tagger`.

            .. code-block:: python

                [
                    ( 0, 2, 'GPE', '美國', ),   # name-entity 1
                    ( 3, 5, 'ORG', '參議院', ), # name-entity 2
                ]
    """

    item_class = NerToken

    to_text = NotImplemented
    from_text = NotImplemented

    ########################################################################################################################

    @classmethod
    def from_tagger(cls, data):
        """Construct an instance from CkipTagger format."""
        return cls(map(cls.item_class.from_tagger, data))

    def to_tagger(self):
        """Transform to CkipTagger format."""
        return list(item.to_tagger() for item in self)

################################################################################################################################

class NerParagraph(_BaseList):
    """A list of named-entity recognition sentence.

    .. admonition:: Data Structure Examples

        Text format
            Not implemented

        List format
            Used for :meth:`from_list` and :meth:`to_list`.

            .. code-block:: python

                [
                    [ # Sentence 1
                        [ '中文字', 'LANGUAGE', (0, 3), ],
                    ],
                    [ # Sentence 2
                        [ '美國', 'GPE', (0, 2), ],
                        [ '參議院', 'ORG', (3, 5), ],
                    ],
                ]

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.

            .. code-block:: python

                [
                    [ # Sentence 1
                        { 'word': '中文字', 'ner': 'LANGUAGE', 'idx': (0, 3), },
                    ],
                    [ # Sentence 2
                        { 'word': '美國', 'ner': 'GPE', 'idx': (0, 2), },
                        { 'word': '參議院', 'ner': 'ORG', 'idx': (3, 5), },
                    ],
                ]

        CkipTagger format
            Used for :meth:`from_tagger` and :meth:`to_tagger`.

            .. code-block:: python

                [
                    [ # Sentence 1
                        ( 0, 3, 'LANGUAGE', '中文字', ),
                    ],
                    [ # Sentence 2
                        ( 0, 2, 'GPE', '美國', ),
                        ( 3, 5, 'ORG', '參議院', ),
                    ],
                ]
    """

    item_class = NerSentence

    to_text = NotImplemented
    from_text = NotImplemented

    ########################################################################################################################

    @classmethod
    def from_tagger(cls, data):
        """Construct an instance from CkipTagger format."""
        return cls(map(cls.item_class.from_tagger, data))

    def to_tagger(self):
        """Transform to CkipTagger format."""
        return list(item.to_tagger() for item in self)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides built-in sentence segmentation driver.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

import re as _re

from ckipnlp.container import (
    TextParagraph as _TextParagraph,
)

from .base import (
    BaseDriver as _BaseDriver,
)

################################################################################################################################

class CkipSentenceSegmenter(_BaseDriver):  # pylint: disable=too-few-public-methods
    """The CKIP sentence segmentation driver.

    Arguments
    ---------
        lazy : bool
            Lazy initialize the driver.
        delims : str
            The delimiters.
        keep_delims : bool
            Keep the delimiters.

    .. method:: __call__(*, raw, keep_all=True)

        Apply sentence segmentation.

        Parameters
            **raw** (*str*) — The raw text.

        Returns
            **text** (:class:`TextParagraph <ckipnlp.container.text.TextParagraph>`) — The sentences.
    """

    driver_type = 'sentence_segmenter'
    driver_family = 'default'
    driver_inputs = ('raw',)

    def __init__(self, *, lazy=False, delims='\n', keep_delims=False):
        super().__init__(lazy=lazy)

        self.delims = delims
        self._keep_delims = keep_delims

    def _init(self):
        pass

    def _call(self, *, raw):
        assert isinstance(raw, str)

        if not self._keep_delims:
            # Replace spaces
            text = _re.sub(rf'[^\S{self.delims}]', '', raw)

            # Segment
            text = _re.split(rf'[{self.delims}]+', text)

            # Remove empty lines
            text = filter(None, text)

        else:
            text = _re.split(rf'([{self.delims}]+)', raw)
            if text[-1] == '':
                del text[-1]
            if len(text) % 2:
                text.append('')
            text = [word+punct for word, punct in zip(text[::2], text[1::2])]

        return _TextParagraph.from_text(text)



from ._alumnus import ALUMNUS_WORDS
from ._aristocrat import ARISTOCRAT_WORDS
from ._candidate import CANDIDATE_WORDS
from ._fans import FANS_WORDS
from ._humanized import HUMANIZED_WORDS
from ._member import MEMBER_WORDS
from ._owner import OWNER_WORDS
from ._professional import PROFESSIONAL_WORDS
from ._recipient import RECIPIENT_WORDS
from ._religious_follower import RELIGIOUS_FOLLOWER_WORDS
from ._sage import SAGE_WORDS
from ._sick_patient import SICK_PATIENT_WORDS
from ._student import STUDENT_WORDS
from ._talent import TALENT_WORDS
from ._victim import VICTIM_WORDS
from ._warrior import WARRIOR_WORDS

from ._instance import HUMAN_INSTANCE_WORDS

HUMAN_WORDS = {
    *ALUMNUS_WORDS,
    *ARISTOCRAT_WORDS,
    *CANDIDATE_WORDS,
    *FANS_WORDS,
    *HUMANIZED_WORDS,
    *MEMBER_WORDS,
    *OWNER_WORDS,
    *PROFESSIONAL_WORDS,
    *RECIPIENT_WORDS,
    *RELIGIOUS_FOLLOWER_WORDS,
    *SAGE_WORDS,
    *SICK_PATIENT_WORDS,
    *STUDENT_WORDS,
    *TALENT_WORDS,
    *VICTIM_WORDS,
    *WARRIOR_WORDS,

    *HUMAN_INSTANCE_WORDS,
}



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# pylint: disable=missing-docstring

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

################################################################################################################################

SUBJECT_ROLES = {'agent', 'causer', 'experiencer'}
OBJECT_ROLES = {'benefactor', 'companion', 'comparison', 'goal', 'range', 'source', 'target'}
NEUTRAL_ROLES = {'theme', 'topic'}
APPOSITION_ROLES = {'apposition'}

HUMAN_ROLES = {
    *SUBJECT_ROLES,
    *OBJECT_ROLES,
    *NEUTRAL_ROLES,
    *APPOSITION_ROLES,
}



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides built-in coreference resolution driver.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

import numpy as _np

from treelib import (
    Tree as _Tree,
)

from ckipnlp.container import (
    TextParagraph as _TextParagraph,
    SegParagraph as _SegParagraph,
    ParseParagraph as _ParseParagraph,
    NerParagraph as _NerParagraph,
    CorefToken as _CorefToken,
    CorefSentence as _CorefSentence,
    CorefParagraph as _CorefParagraph,
)

from ckipnlp.data.conparse import (
    APPOSITION_ROLES as _APPOSITION_ROLES,
)

from ckipnlp.data.coref import (
    HUMAN_WORDS as _HUMAN_WORDS,
    PRONOUN_3RD_WORDS as _PRONOUN_3RD_WORDS,
    SELF_WORDS as _SELF_WORDS,
)

from .base import (
    BaseDriver as _BaseDriver,
)

################################################################################################################################

class CkipCorefChunker(_BaseDriver):  # pylint: disable=too-few-public-methods
    """The CKIP coreference resolution driver.

    Arguments
    ---------
        lazy : bool
            Lazy initialize the driver.

    .. method:: __call__(*, conparse)

        Apply coreference delectation.

        Parameters
            **conparse** (:class:`~ckipnlp.container.parse.ParseParagraph`) — The constituency-parsing sentences.

        Returns
            **coref** (:class:`~ckipnlp.container.coref.CorefParagraph`) — The coreference results.
    """

    driver_type = 'coref_chunker'
    driver_family = 'default'
    driver_inputs = None

    def _init(self):
        pass

    def _call(self, *, conparse):
        assert isinstance(conparse, _ParseParagraph)

        # Convert to tree structure
        tree_list = [
            [
                (clause.to_tree(), clause.delim,)
                for clause in sent
            ] for sent in conparse
        ]

        # Find coreference
        coref_tree = self._get_coref(tree_list)

        # # Get results
        coref = self._get_result(tree_list, coref_tree=coref_tree)

        return coref

    ########################################################################################################################

    @classmethod
    def _get_coref(cls, tree_list):

        coref_tree = _Tree()
        coref_tree.create_node(identifier=0)
        dummy_id = (-1, -1, -1)
        coref_tree.create_node(identifier=dummy_id, parent=0, data=True)

        name2node = {}  # name => ((sent_id, clause_id,), (node_id, is_prepend))

        curr_source = None    # the current coref source
        last_source = None    # the last coref source
        last_subject = None   # the last coref subject

        last_sent_pos = None  # the POS-tag of last sentence

        # Find coref
        for sent_id, clause_list in enumerate(tree_list):
            for clause_id, (tree, _) in enumerate(clause_list):
                tree_key = (sent_id, clause_id,)

                if tree is None:
                    continue
                curr_sent_pos = tree[tree.root].data.pos

                # Get relations
                appositions = []
                for rel in tree.get_relations():
                    if rel.relation.data.role in _APPOSITION_ROLES:
                        appositions.append(((rel.head.identifier, False), (rel.tail.identifier, False),))

                # Get sources/targets
                node_keys = {}
                for node_key in cls._get_sources(tree): # Source
                    node_keys[node_key] = True
                for node_key in cls._get_targets(tree): # Target
                    node_keys[node_key] = False
                subject_keys = set(cls._get_subjects(tree)) # Subject

                for node_key, ntype in sorted(node_keys.items()):
                    if ntype: # Assign ref_id to sources
                        source = tree[node_key[0]]
                        curr_source = (tree_key, node_key,)

                        parent_id = name2node.get(source.data.word, None)
                        if parent_id:
                            coref_tree.create_node(identifier=(tree_key, node_key,), parent=parent_id, data=True)
                        else:
                            name2node[source.data.word] = (tree_key, node_key,)
                            coref_tree.create_node(identifier=(tree_key, node_key,), parent=0, data=True)

                    else: # Link targets to previous sources
                        if node_key[1] and last_subject:
                            coref_tree.create_node(identifier=(tree_key, node_key,), parent=last_subject, data=False)
                        if not node_key[1]:
                            if curr_source and tree[node_key[0]].data.word in _SELF_WORDS:
                                coref_tree.create_node(identifier=(tree_key, node_key,), parent=curr_source, data=False)
                            elif last_source:
                                coref_tree.create_node(identifier=(tree_key, node_key,), parent=last_source, data=False)
                            else:
                                coref_tree.create_node(identifier=(tree_key, node_key,), parent=dummy_id, data=False)

                # Merge apposition (apposition role)
                for head_key, tail_key in appositions:
                    head_key = (tree_key, head_key,)
                    tail_key = (tree_key, tail_key,)

                    if coref_tree.contains(head_key) and coref_tree.contains(tail_key):
                        if coref_tree.is_ancestor(head_key, tail_key) or \
                           coref_tree.is_ancestor(tail_key, head_key):
                            continue

                        if coref_tree[head_key].data:  # Head is a source
                            coref_tree.move_node(tail_key, head_key)
                        elif coref_tree[tail_key].data:  # Tail is a source
                            coref_tree.move_node(head_key, tail_key)
                        else:
                            coref_tree.move_node(tail_key, head_key)

                # Merge apposition (NP sentences)
                if curr_sent_pos == 'NP' and last_sent_pos == 'NP' and last_subject:
                    for node_key, ntype in node_keys.items():
                        if ntype: # Merge sources only
                            source_id = (tree_key, node_key,)
                            if coref_tree.contains(source_id):
                                coref_tree.move_node(source_id, last_subject)

                # Update subject
                if curr_sent_pos in ('NP', 'S'):
                    last_subject = None
                    for node_key, ntype in sorted(node_keys.items(), key=lambda x: x[::-1]):
                        if node_key in subject_keys:
                            last_subject = (tree_key, node_key,)
                            break

                # Update last
                last_source = curr_source
                last_sent_pos = curr_sent_pos

        # Remove dummy node
        coref_tree.remove_node(dummy_id)

        return coref_tree


    @classmethod
    def _get_result(cls, tree_list, *, coref_tree):

        # Assign coref ID
        node2coref = {}  # ((sent_id, clause_id,), (node_id, is_prepend)) => ref_id
        sources = set()  # ((sent_id, clause_id,), (node_id, is_prepend))

        for ref_id, coref_source in enumerate(coref_tree.children(coref_tree.root)):
            tree_key, node_key = coref_source.identifier
            sources.add((tree_key, node_key,))
            for tree_key, node_key in coref_tree.expand_tree(coref_source.identifier):
                node2coref[tree_key, node_key] = ref_id

        # Generate result
        tokens_list = _CorefParagraph()
        for sent_id, clause_list in enumerate(tree_list):
            tokens = _CorefSentence()
            tokens_list.append(tokens)

            for clause_id, (tree, delim,) in enumerate(clause_list):
                tree_key = (sent_id, clause_id,)

                if tree is not None:
                    nodes = tree.leaves()

                    for node in nodes:
                        key = (tree_key, (node.identifier, True),)
                        ref_id = node2coref.get(key, -1)
                        if ref_id >= 0:
                            tokens.append(_CorefToken(  # pylint: disable=no-value-for-parameter
                                word=None,
                                idx=(clause_id, None,),
                                coref=(ref_id, 'zero'),
                            ))

                        key = (tree_key, (node.identifier, False),)
                        ref_id = node2coref.get(key, -1)
                        if ref_id >= 0:
                            tokens.append(_CorefToken(  # pylint: disable=no-value-for-parameter
                                word=node.data.word,
                                idx=(clause_id, node.identifier,),
                                coref=(ref_id, 'source' if key in sources else 'target',),
                            ))
                        else:
                            tokens.append(_CorefToken(  # pylint: disable=no-value-for-parameter
                                word=node.data.word,
                                idx=(clause_id, node.identifier,),
                                coref=None,
                            ))
                if delim:
                    tokens.append(_CorefToken(  # pylint: disable=no-value-for-parameter
                        word=delim,
                        idx=(clause_id, None,),
                        coref=None,
                    ))

        return tokens_list

    ########################################################################################################################

    @staticmethod
    def transform_ws(*, text, ws, ner):
        """Transform word-segmented sentence lists (create a new instance)."""
        assert isinstance(text, _TextParagraph)
        assert isinstance(ws, _SegParagraph)
        assert isinstance(ner, _NerParagraph)

        ws_new = []
        for line, line_ws, line_ner in zip(text, ws, ner):
            line_bi = _np.zeros(len(line)+1, dtype=_np.bool)
            line_bi[0] = True
            line_bi[_np.cumsum(list(map(len, line_ws)))] = True
            for _, _, (idx0, idx1,) in line_ner:
                line_bi[[idx0, idx1]] = True
                line_bi[idx0+1:idx1] = False
            idxs = _np.where(line_bi)[0]
            ws_new.append([line[idx0:idx1] for idx0, idx1 in zip(idxs[:-1], idxs[1:])])
        return _SegParagraph.from_list(ws_new)

    @staticmethod
    def transform_pos(*, ws, pos, ner):
        """Transform pos-tag sentence lists (modify in-place)."""
        assert isinstance(ws, _SegParagraph)
        assert isinstance(pos, _SegParagraph)
        assert isinstance(ner, _NerParagraph)

        for line_ws, line_pos, line_ner in zip(ws, pos, ner):
            idxmap = {idx: i for i, idx in enumerate(_np.cumsum(list(map(len, line_ws))))}
            for token in line_ner:
                if token.ner == 'PERSON':
                    line_pos[idxmap[token.idx[1]]] = 'Nb'

    ########################################################################################################################

    @classmethod
    def _get_sources(cls, tree):
        """Get sources of a tree

        Parameters
        ----------
            tree : :class:`~ckipnlp.container.util.parse_tree.ParseTree`
                the constituency parsing tree.

        Yields
        ------
            int
                the identifier of source nodes.
            bool
                prepend this node or not.

        Notes
        -----
            A node can be a coreference source if either:

            1. POS-tag is `Nb`
            2. is one of the human words from E-HowNet

        """
        for node in tree.leaves():
            if cls._is_human_word(node):
                yield node.identifier, False

    @classmethod
    def _get_targets(cls, tree):
        """Get targets of a tree

        Parameters
        ----------
            tree : :class:`~ckipnlp.container.util.parse_tree.ParseTree`
                the constituency parsing tree.

        Yields
        ------
            int
                the identifier of source nodes.
            bool
                prepend this node or not.

        Notes
        -----
            A node can be a coreference target if either:

            1. POS-tag is `Nh`
            2. is one of the pronoun words from E-HowNet

        """
        root = tree[tree.root]
        leaves = tree.leaves()

        if root.data.pos == 'VP':
            if leaves[0].data.pos.startswith('Cb'): # coref will be inserted after this Cb node
                yield leaves[1].identifier, True
            else: # coref will be inserted in front of the whole sentence
                yield leaves[0].identifier, True

        for node in leaves:
            if cls._is_pronoun_word(node):
                yield node.identifier, False

    ########################################################################################################################

    @classmethod
    def _get_subjects(cls, tree):
        """Get subjects of a tree

        Parameters
        ----------
            tree : :class:`~ckipnlp.container.util.parse_tree.ParseTree`
                the constituency parsing tree.

        Yields
        ------
            int
                the identifier of subject node.
            bool
                prepend this node or not.
        """
        for node in tree.get_subjects():
            yield node.identifier, False

    ########################################################################################################################

    @staticmethod
    def _is_human_word(node):
        return node.data.pos.startswith('Nb') or node.data.pos.startswith('N') and node.data.word in _HUMAN_WORDS

    @staticmethod
    def _is_pronoun_word(node):
        return node.data.pos.startswith('Nh') or node.data.pos.startswith('N') and node.data.word in _PRONOUN_3RD_WORDS



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides containers for word-segmented sentences.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from .base import (
    BaseList as _BaseList,
    BaseSentence0 as _BaseSentence0,
)

################################################################################################################################

class SegSentence(_BaseSentence0):
    """A word-segmented sentence.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`from_text` and :meth:`to_text`.

            .. code-block:: python

                '中文字\u3000耶\u3000，\u3000啊\u3000哈\u3000哈哈\u3000。' # Words segmented by \\u3000 (full-width space)

        List/Dict format
            Used for :meth:`from_list`, :meth:`to_list`, :meth:`from_dict`, and :meth:`to_dict`.

            .. code-block:: python

                [ '中文字', '耶', '，', '啊', '哈', '哈哈', '。', ]

    .. note::

        This class is also used for part-of-speech tagging.

    """

    item_class = str

class SegParagraph(_BaseList):
    """A list of word-segmented sentences.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`from_text` and :meth:`to_text`.

            .. code-block:: python

                [
                    '中文字\u3000耶\u3000，\u3000啊\u3000哈\u3000哈\u3000。',        # Sentence 1
                    '「\u3000完蛋\u3000了\u3000！\u3000」\u3000，\u3000畢卡索\u3000他\u3000想', # Sentence 2
                ]

        List/Dict format
            Used for :meth:`from_list`, :meth:`to_list`, :meth:`from_dict`, and :meth:`to_dict`.

            .. code-block:: python

                [
                    [ '中文字', '耶', '，', '啊', '哈', '哈哈', '。', ],            # Sentence 1
                    [ '「', '完蛋', '了', '！', '」', '，', '畢卡索', '他', '想', ], # Sentence 2
                ]

    .. note::

        This class is also used for part-of-speech tagging.

    """

    item_class = SegSentence



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides base drivers.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from abc import (
    ABCMeta as _ABCMeta,
    abstractmethod as _abstractmethod,
)

from ckipnlp.util.logger import (
    get_logger as _get_logger,
)

################################################################################################################################

class DriverRegister:
    """The driver registering utility."""

    _DRIVERS = {}

    @staticmethod
    def get(driver_type, driver_family):  # pylint: disable=missing-docstring
        if driver_family is None:
            return DummyDriver

        driver = DriverRegister._DRIVERS.get((driver_type, driver_family,))
        if not driver:
            raise KeyError(f'{driver_type} is not implemented for type {driver_family}')
        if not driver.is_dummy:
            _get_logger().debug(f'Use {driver.__name__} ...')

        return driver

################################################################################################################################

class BaseDriver(metaclass=_ABCMeta):
    """The base CKIPNLP driver."""

    is_dummy = False

    def __init__(self, *, lazy=False):
        self._core = None
        self._inited = False

        if not lazy:
            self.init()

    def init(self):  # pylint: disable=missing-docstring
        if self._inited:
            return
        _get_logger().info(f'Initializing {self.__class__.__name__} ...')
        self._init()
        self._inited = True

    def __call__(self, *args, **kwargs):
        self.init()
        return self._call(*args, **kwargs)

    ########################################################################################################################

    @_abstractmethod
    def driver_type(self):  # pylint: disable=missing-docstring
        return NotImplemented

    @_abstractmethod
    def driver_family(self):  # pylint: disable=missing-docstring
        return NotImplemented

    @_abstractmethod
    def driver_inputs(self):  # pylint: disable=missing-docstring
        return NotImplemented

    @_abstractmethod
    def _init(self):
        return NotImplemented

    @_abstractmethod
    def _call(self):
        return NotImplemented

    def _call_from_pipeline(self, pipeline, doc):
        return self.__call__(**{
            key: pipeline._get(key, doc) for key in self.driver_inputs  # pylint: disable=protected-access
        })

    ########################################################################################################################

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        driver_type = cls.driver_type
        driver_family = cls.driver_family

        key = (driver_type, driver_family,)
        assert key not in DriverRegister._DRIVERS, f'{key} already registered!'  # pylint: disable=protected-access
        DriverRegister._DRIVERS[key] = cls  # pylint: disable=protected-access

################################################################################################################################

class DummyDriver(BaseDriver):
    """The dummy driver."""

    driver_type = None
    driver_family = None
    driver_inputs = None
    is_dummy = True

    def __init__(self, *, lazy=False):
        super().__init__(lazy=lazy)
        self._inited = True

    def _init(self):
        pass

    def _call(self):
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides kernel CKIPNLP pipeline.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from collections.abc import (
    Mapping as _Mapping,
)

from ckipnlp.driver.base import (
    DriverRegister as _DriverRegister,
)

###############################################################################################################################)

class CkipDocument(_Mapping):
    """The kernel document.

    Attributes
    ----------
        raw : str
            The unsegmented text input.
        text : :class:`~ckipnlp.container.text.TextParagraph`
            The sentences.
        ws : :class:`~ckipnlp.container.seg.SegParagraph`
            The word-segmented sentences.
        pos : :class:`~ckipnlp.container.seg.SegParagraph`
            The part-of-speech sentences.
        ner : :class:`~ckipnlp.container.ner.NerParagraph`
            The named-entity recognition results.
        conparse : :class:`~ckipnlp.container.parse.ParseParagraph`
            The constituency-parsing sentences.
    """

    __keys = ('raw', 'text', 'ws', 'pos', 'ner', 'conparse',)

    def __init__(self, *, raw=None, text=None, ws=None, pos=None, ner=None, conparse=None):
        self.raw = raw
        self.text = text
        self.ws = ws
        self.pos = pos
        self.ner = ner
        self.conparse = conparse

        self._wspos = None

    def __len__(self):
        return len(self.__keys)

    def __iter__(self):
        yield from self.__keys

    def __getitem__(self, key):
        return getattr(self, key)

################################################################################################################################

class CkipPipeline:
    """The kernel pipeline.

    Arguments
    ---------
        sentence_segmenter : str
            The type of sentence segmenter.

        word_segmenter : str
            The type of word segmenter.

        pos_tagger : str
            The type of part-of-speech tagger.

        ner_chunker : str
            The type of named-entity recognition chunker.

        con_parser : str
            The type of constituency parser.

    Other Parameters
    ----------------
        lazy : bool
            Lazy initialize the drivers.

        opts : Dict[str, Dict]
            The driver options. Key: driver name (e.g. `'sentence_segmenter'`); Value: a dictionary of options.
    """

    def __init__(self, *,
            sentence_segmenter='default',
            word_segmenter='tagger',
            pos_tagger='tagger',
            con_parser='classic-client',
            ner_chunker='tagger',
            lazy=True,
            opts={},
        ):

        if word_segmenter == '_classic':
            word_segmenter = 'classic'
        if pos_tagger == '_classic':
            pos_tagger = 'classic'

        # WS & POS
        if pos_tagger == 'classic':
            assert word_segmenter == 'classic', 'CkipClassicPosTagger must be used with CkipClassicWordSegmenter together!'
            self._wspos_driver = _DriverRegister.get('_wspos', '_classic')(
                lazy=lazy, **opts.get('word_segmenter', {}), **opts.get('pos_tagger', {}),
            )
            word_segmenter = '_classic'
            pos_tagger = '_classic'
        else:
            self._wspos_driver = _DriverRegister.get(None, None)()

        self._sentence_segmenter = _DriverRegister.get('sentence_segmenter', sentence_segmenter)(
            lazy=lazy, **opts.get('sentence_segmenter', {}),
        )
        self._word_segmenter = _DriverRegister.get('word_segmenter', word_segmenter)(
            lazy=lazy, **opts.get('word_segmenter', {}),
        )
        self._pos_tagger = _DriverRegister.get('pos_tagger', pos_tagger)(
            lazy=lazy, **opts.get('pos_tagger', {}),
        )
        self._con_parser = _DriverRegister.get('con_parser', con_parser)(
            lazy=lazy, **opts.get('con_parser', {}),
        )
        self._ner_chunker = _DriverRegister.get('ner_tagger', ner_chunker)(
            lazy=lazy, **opts.get('ner_chunker', {}),
        )

    ########################################################################################################################

    def _get(self, key, doc):
        driver, name = {
            'raw': (
                None, None,
            ),
            '_wspos': (
                self._wspos_driver, 'classic word segmentation',
            ),
            'text': (
                self._sentence_segmenter, 'sentence segmentation',
            ),
            'ws': (
                self._word_segmenter, 'word segmentation',
            ),
            'pos': (
                self._pos_tagger, 'part-of-speech tagging',
            ),
            'conparse': (
                self._con_parser, 'constituency parsing',
            ),
            'ner': (
                self._ner_chunker, 'named-entity recognition',
            ),
        }[key]

        if doc[key] is NotImplemented:
            raise RecursionError('Loop dependence detected!')

        if doc[key] is None:
            setattr(doc, key, NotImplemented)

            if key == 'raw':
                raise AttributeError('No raw text!')
            elif not driver.is_dummy:
                ret = driver._call_from_pipeline(self, doc)  # pylint: disable=protected-access
                setattr(doc, key, ret)
            else:
                raise AttributeError(f'No {name} driver / no {name} as input!')

        return doc[key]

    ########################################################################################################################

    def get_text(self, doc):
        """Apply sentence segmentation.

        Arguments
        ---------
            doc : :class:`CkipDocument`
                The input document.

        Returns
        -------
            doc.text : :class:`~ckipnlp.container.text.TextParagraph`
                The sentences.

        .. note::

            This routine modify **doc** inplace.
        """
        return self._get('text', doc)

    ########################################################################################################################

    def get_ws(self, doc):
        """Apply word segmentation.

        Arguments
        ---------
            doc : :class:`CkipDocument`
                The input document.

        Returns
        -------
            doc.ws : :class:`~ckipnlp.container.seg.SegParagraph`
                The word-segmented sentences.

        .. note::

            This routine modify **doc** inplace.
        """
        return self._get('ws', doc)

    ########################################################################################################################

    def get_pos(self, doc):
        """Apply part-of-speech tagging.

        Arguments
        ---------
            doc : :class:`CkipDocument`
                The input document.

        Returns
        -------
            doc.pos : :class:`~ckipnlp.container.seg.SegParagraph`
                The part-of-speech sentences.

        .. note::

            This routine modify **doc** inplace.
        """
        return self._get('pos', doc)

    ########################################################################################################################

    def get_ner(self, doc):
        """Apply named-entity recognition.

        Arguments
        ---------
            doc : :class:`CkipDocument`
                The input document.

        Returns
        -------
            doc.ner : :class:`~ckipnlp.container.ner.NerParagraph`
                The named-entity recognition results.

        .. note::

            This routine modify **doc** inplace.
        """
        return self._get('ner', doc)

    ########################################################################################################################

    def get_conparse(self, doc):
        """Apply constituency parsing.

        Arguments
        ---------
            doc : :class:`CkipDocument`
                The input document.

        Returns
        -------
            doc.conparse : :class:`~ckipnlp.container.parse.ParseParagraph`
                The constituency parsing sentences.

        .. note::

            This routine modify **doc** inplace.
        """
        return self._get('conparse', doc)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides tree containers for parsed sentences.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'


from collections import (
    deque as _deque,
)

from typing import (
    NamedTuple as _NamedTuple,
)

from treelib import (
    Tree as _Tree,
    Node as _Node,
)

from ckipnlp.data.conparse import (
    SUBJECT_ROLES as _SUBJECT_ROLES,
    NEUTRAL_ROLES as _NEUTRAL_ROLES,
)

from ..base import (
    Base as _Base,
    BaseTuple as _BaseTuple,
)

################################################################################################################################

class _ParseNodeData(_NamedTuple):
    role: str = None
    pos: str = None
    word: str = None

class ParseNodeData(_BaseTuple, _ParseNodeData):
    """A parse node.

    Attributes
    ----------
        role : str
            the semantic role.
        pos : str
            the POS-tag.
        word : str
            the text term.

    Note
    ----
        This class is an subclass of :class:`tuple`. To change the attribute, please create a new instance instead.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`from_text` and :meth:`to_text`.

            .. code-block:: python

                'Head:Na:中文字'  # role / POS-tag / text-term

        List format
            Not implemented.

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.

            .. code-block:: python

                {
                    'role': 'Head',   # role
                    'pos': 'Na',      # POS-tag
                    'word': '中文字',  # text term
                }
    """

    from_list = NotImplemented
    to_list = NotImplemented

    @classmethod
    def from_text(cls, data):
        """Construct an instance from text format.

        Parameters
        ----------
            data : str
                text such as ``'Head:Na:中文字'``.

        .. note::

            - ``'Head:Na:中文字'`` -> **role** = ``'Head'``, **pos** = ``'Na'``, **word** = ``'中文字'``
            - ``'Head:Na'``       -> **role** = ``'Head'``, **pos** = ``'Na'``, **word** = ``None``
            - ``'Na'``            -> **role** = ``None``,   **pos** = ``'Na'``, **word** = ``None``
        """
        if ':' in data:
            fields = data.split(':')
            return cls(*fields)
        return cls(pos=data)  # pylint: disable=no-value-for-parameter

    def to_text(self):
        return ':'.join(filter(None, self))

################################################################################################################################

class ParseNode(_Base, _Node):
    """A parse node for tree.

    Attributes
    ----------
        data : :class:`ParseNodeData`

    See Also
    --------
        treelib.tree.Node: Please refer `<https://treelib.readthedocs.io/>`__ for built-in usages.

    .. admonition:: Data Structure Examples

        Text format
            Not implemented.

        List format
            Not implemented.

        Dict format
            Used for :meth:`to_dict`.

            .. code-block:: python

                {
                    'role': 'Head',   # role
                    'pos': 'Na',      # POS-tag
                    'word': '中文字',  # text term
                }
    """

    data_class = ParseNodeData

    from_dict = NotImplemented

    from_text = NotImplemented
    to_text = NotImplemented
    from_list = NotImplemented
    to_list = NotImplemented

    def __repr__(self):
        return '{name}(tag={tag}, identifier={identifier})'.format(
            name=self.__class__.__name__,
            tag=self.tag,
            identifier=self.identifier,
        )

    def to_dict(self):
        return dict(id=self.identifier, data=self.data.to_dict())

################################################################################################################################

class _ParseRelation(_NamedTuple):
    head: ParseNode
    tail: ParseNode
    relation: ParseNode

class ParseRelation(_Base, _ParseRelation):
    """A parse relation.

    Attributes
    ----------
        head : :class:`ParseNode`
            the head node.
        tail : :class:`ParseNode`
            the tail node.
        relation : :class:`ParseNode`
            the relation node. (the semantic role of this node is the relation.)

    Notes
    -----
        The parent of the relation node is always the common ancestor of the head node and tail node.

    .. admonition:: Data Structure Examples

        Text format
            Not implemented.

        List format
            Not implemented.

        Dict format
            Used for :meth:`to_dict`.

            .. code-block:: python

                {
                    'tail': { 'role': 'Head', 'pos': 'Nab', 'word': '中文字' }, # head node
                    'tail': { 'role': 'particle', 'pos': 'Td', 'word': '耶' }, # tail node
                    'relation': 'particle',  # relation
                }
    """

    from_dict = NotImplemented

    from_text = NotImplemented
    to_text = NotImplemented
    from_list = NotImplemented
    to_list = NotImplemented

    def __repr__(self):
        ret = '{name}(head=({head}, {head_id}), tail=({tail}, {tail_id}), relation=({rel}, {rel_id}))' if self.head_first \
         else '{name}(tail=({tail}, {tail_id}), head=({head}, {head_id}), relation=({rel}, {rel_id}))'
        return ret.format(
            name=type(self).__name__,
            head=self.head.tag,
            head_id=self.head.identifier,
            tail=self.tail.tag,
            tail_id=self.tail.identifier,
            rel=self.relation.data.role,
            rel_id=self.relation.identifier,
        )

    @property
    def head_first(self):  # pylint: disable=missing-docstring
        return self.head.identifier <= self.tail.identifier

    def to_dict(self):
        return dict(head=self.head.to_dict(), tail=self.tail.to_dict(), relation=self.relation.data.role)

################################################################################################################################

class ParseTree(_Base, _Tree):
    """A parse tree.

    See Also
    --------
        treereelib.tree.Tree: Please refer `<https://treelib.readthedocs.io/>`__ for built-in usages.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`from_text` and :meth:`to_text`.

            .. code-block:: python

                'S(Head:Nab:中文字|particle:Td:耶)'

        List format
            Not implemented.

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.
            A dictionary such as ``{ 'id': 0, 'data': { ... }, 'children': [ ... ] }``,
            where ``'data'`` is a dictionary with the same format as :meth:`ParseNodeData.to_dict`,
            and ``'children'`` is a list of dictionaries of subtrees with the same format as this tree.

            .. code-block:: python

                {
                    'id': 0,
                    'data': {
                        'role': None,
                        'pos': 'S',
                        'word': None,
                    },
                    'children': [
                        {
                            'id': 1,
                            'data': {
                                'role': 'Head',
                                'pos': 'Nab',
                                'word': '中文字',
                            },
                            'children': [],
                        },
                        {
                            'id': 2,
                            'data': {
                                'role': 'particle',
                                'pos': 'Td',
                                'word': '耶',
                            },
                            'children': [],
                        },
                    ],
                }

        Penn Treebank format
            Used for :meth:`from_penn` and :meth:`to_penn`.

            .. code-block:: python

                [
                    'S',
                    [ 'Head:Nab', '中文字', ],
                    [ 'particle:Td', '耶', ],
                ]

    .. note::

        One may use :meth:`to_penn` together with `SvgLing <https://pypi.org/project/svgling/>`__ to generate SVG tree graphs.

    """

    node_class = ParseNode

    from_list = NotImplemented
    to_list = NotImplemented

    def __str__(self):
        return self.to_text()

    ########################################################################################################################

    @classmethod
    def from_text(cls, data):
        """Construct an instance from text format.

        Parameters
        ----------
            data : str
                A parse tree in text format (:class:`ParseClause.clause <.parse.ParseClause>`).

        .. seealso::
            :meth:`ParseClause.to_tree() <.parse.ParseClause.to_tree>`.
        """

        tree = cls()
        node_id = 0
        node_stack = [None]
        text = ''
        ending = True

        for char in data:
            if char == '(':
                node_data = cls.node_class.data_class.from_text(text)
                tree.create_node(tag=text, identifier=node_id, parent=node_stack[-1], data=node_data)

                node_stack.append(node_id)
                node_id += 1
                text = ''

            elif char == ')':
                if not ending:
                    node_data = cls.node_class.data_class.from_text(text)
                    tree.create_node(tag=text, identifier=node_id, parent=node_stack[-1], data=node_data)
                    node_id += 1

                node_stack.pop()
                text = ''
                ending = True

            elif char == '|':
                if not ending:
                    node_data = cls.node_class.data_class.from_text(text)
                    tree.create_node(tag=text, identifier=node_id, parent=node_stack[-1], data=node_data)
                    node_id += 1

                text = ''
                ending = True

            else:
                ending = False
                text += char

        return tree

    def to_text(self, node_id=None):
        """Transform to plain text.

        Parameters
        ----------
            node_id : int
                Output the plain text format for the subtree under **node_id**.

        Returns
        --------
            str
        """
        if node_id is None:
            node_id = self.root

        node = self[node_id]
        tree_text = node.data.to_text()

        children_text = '|'.join((self.to_text(child.identifier) for child in self.children(node_id)))
        if children_text:
            tree_text = '{}({})'.format(tree_text, children_text)

        return tree_text

    @classmethod
    def from_dict(cls, data):
        """Construct an instance from python built-in containers.

        Parameters
        ----------
            data : str
                A parse tree in dictionary format.
        """
        tree = cls()

        node_queue = _deque()
        node_queue.append((data, None,))

        while node_queue:
            node_dict, parent_id = node_queue.popleft()
            node_id = node_dict['id']
            node_data = cls.node_class.data_class.from_dict(node_dict['data'])
            tree.create_node(tag=node_data.to_text(), identifier=node_id, parent=parent_id, data=node_data)

            for child in node_dict['children']:
                node_queue.append((child, node_id,))

        return tree

    def to_dict(self, node_id=None):
        """Transform to python built-in containers.

        Parameters
        ----------
            node_id : int
                Output the plain text format for the subtree under **node_id**.

        Returns
        -------
            str
        """
        if node_id is None:
            node_id = self.root

        tree_dict = self[node_id].to_dict()
        tree_dict['children'] = []

        for child in self.children(node_id):
            tree_dict['children'].append(self.to_dict(child.identifier))

        return tree_dict

    @classmethod
    def from_penn(cls, data):
        """Construct an instance from Penn Treebank format."""
        tree = cls()

        node_stack = _deque()
        node_stack.append((data, None,))

        node_id = 0

        while node_stack:
            penn_data, parent_id = node_stack.pop()

            if not penn_data:
                raise SyntaxError(f'Empty node #{node_id}')

            if not isinstance(penn_data[0], str):
                raise SyntaxError(f'First element of a node must be string, got {type(penn_data[0])}')

            if len(penn_data) == 2 and isinstance(penn_data[-1], str):
                penn_data = (':'.join(penn_data),)

            node_data = cls.node_class.data_class.from_text(penn_data[0])
            tree.create_node(tag=node_data.to_text(), identifier=node_id, parent=parent_id, data=node_data)

            for child in penn_data[-1:0:-1]:
                node_stack.append((child, node_id,))
            node_id += 1

        return tree

    def to_penn(self, node_id=None, *, with_role=True, with_word=True, sep=':'):
        """Transform to Penn Treebank format.

        Parameters
        ----------
            node_id : int
                Output the plain text format for the subtree under **node_id**.
            with_role : bool
                Contains role-tag or not.
            with_word : bool
                Contains word or not.
            sep : str
                The seperator between role and POS-tag.

        Returns
        -------
            list
        """
        if node_id is None:
            node_id = self.root
        node = self[node_id]

        penn_data = [f'{node.data.role}{sep}{node.data.pos}' if with_role and node.data.role else node.data.pos,]
        if with_word and node.data.word:
            penn_data.append(node.data.word)

        for child in self.children(node_id):
            penn_data.append(self.to_penn(child.identifier, with_role=with_role, with_word=with_word, sep=sep))

        return penn_data

    ########################################################################################################################

    def show(self, *,
        key=lambda node: node.identifier,
        idhidden=False,
        **kwargs,
    ):
        """Show pretty tree."""
        _Tree.show(self, key=key, idhidden=idhidden, **kwargs)

    def get_children(self, node_id, *, role):
        """Get children of a node with given role.

        Parameters
        ----------
            node_id : int
                ID of target node.
            role : str
                the target role.

        Yields
        ------
            :class:`ParseNode`
                the children nodes with given role.
        """
        for child in self.children(node_id):
            if child.data.role == role:
                yield child

    def get_heads(self, root_id=None, *, semantic=True, deep=True):
        """Get all head nodes of a subtree.

        Parameters
        ----------
            root_id : int
                ID of the root node of target subtree.
            semantic : bool
                use semantic/syntactic policy. For semantic mode, return ``DUMMY`` or ``head`` instead of syntactic ``Head``.
            deep : bool
                find heads recursively.

        Yields
        ------
            :class:`ParseNode`
                the head nodes.
        """
        if root_id is None:
            root_id = self.root

        head_nodes = []
        children = list(self.children(root_id))

        # No child, choose the root node instead
        if not children:
            head_nodes.append(self[root_id])

        # Semantic mode
        if semantic:
            # Find DUMMY
            if not head_nodes:
                for child in children:
                    if child.data.role in ('DUMMY', 'DUMMY1', 'DUMMY2',):
                        head_nodes.append(child)

            # Find head
            if not head_nodes:
                for child in children:
                    if child.data.role == 'head':
                        head_nodes.append(child)

        # Find Head
        if not head_nodes:
            for child in children:
                if child.data.role == 'Head':
                    head_nodes.append(child)

        # Found no head, choose the last child instead
        if not head_nodes:
            head_nodes.append(children[-1])

        # Recursion
        for node in head_nodes:
            if deep and not node.is_leaf():
                yield from self.get_heads(node.identifier, semantic=semantic)
            else:
                yield node

    def get_relations(self, root_id=None, *, semantic=True):
        """Get all relations of a subtree.

        Parameters
        ----------
            root_id : int
                ID of the subtree root node.
            semantic : bool
                please refer :meth:`get_heads` for policy detail.

        Yields
        ------
            :class:`ParseRelation`
                the relations.
        """
        if root_id is None:
            root_id = self.root

        children = list(self.children(root_id))
        head_children = list(self.get_heads(root_id, semantic=semantic, deep=False))

        # Get heads
        for head_node in self.get_heads(root_id, semantic=semantic):
            # Get tails
            for tail in children:
                if tail.data.role != 'Head' and tail not in head_children:
                    if tail.is_leaf():
                        yield ParseRelation(  # pylint: disable=no-value-for-parameter
                            head=head_node, tail=tail, relation=tail,
                        )
                    else:
                        for node in self.get_heads(tail.identifier, semantic=semantic):
                            yield ParseRelation(  # pylint: disable=no-value-for-parameter
                                head=head_node, tail=node, relation=tail,
                            )

        # Recursion
        for child in children:
            yield from self.get_relations(child.identifier, semantic=semantic)

    def get_subjects(self, root_id=None, *, semantic=True, deep=True):
        """Get the subject node of a subtree.

        Parameters
        ----------
            root_id : int
                ID of the root node of target subtree.
            semantic : bool
                please refer :meth:`get_heads` for policy detail.
            deep : bool
                please refer :meth:`get_heads` for policy detail.

        Yields
        ------
            :class:`ParseNode`
                the subject node.

        Notes
        -----
            A node can be a subject if either:

            1. is a head of `NP`
            2. is a head of a subnode (`N`) of `S` with subject role
            3. is a head of a subnode (`N`) of `S` with neutral role and before the head (`V`) of `S`
        """
        if root_id is None:
            root_id = self.root
        root = self[root_id]

        if root.data.pos == 'NP':
            yield from self.get_heads(root.identifier, semantic=semantic, deep=deep)

        elif root.data.pos == 'S':
            for head in self.get_heads(root.identifier, semantic=False, deep=False):
                if head.data.pos.startswith('V'):
                    for subroot in self.children(root.identifier):
                        if subroot.data.pos.startswith('N') and ( \
                            subroot.data.role in _SUBJECT_ROLES or \
                           (subroot.data.role in _NEUTRAL_ROLES and subroot.identifier < head.identifier) \
                        ):
                            yield from self.get_heads(subroot.identifier, semantic=semantic, deep=deep)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module implements CKIPNLP drivers.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from .tagger import (
    CkipTaggerWordSegmenter,
    CkipTaggerPosTagger,
    CkipTaggerNerChunker,
)

from .classic import (
    CkipClassicWordSegmenter,
    CkipClassicConParser,
    CkipClassicConParserClient,
)

from .ss import (
    CkipSentenceSegmenter,
)

from .coref import (
    CkipCorefChunker,
)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides containers for parsed sentences.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from typing import (
    NamedTuple as _NamedTuple,
)

from .base import (
    BaseTuple as _BaseTuple,
    BaseList as _BaseList,
)

from .util.parse_tree import (
    ParseTree as _ParseTree,
)

################################################################################################################################

class _ParseClause(_NamedTuple):
    clause: str = None
    delim: str = ''

class ParseClause(_BaseTuple, _ParseClause):
    """A parse clause.

    Attributes
    ----------
        clause : str
            the parse clause.
        delim : str
            the punctuations after this clause.

    Note
    ----
        This class is an subclass of :class:`tuple`. To change the attribute, please create a new instance instead.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`to_text`.

            .. code-block:: python

                'S(Head:Nab:中文字|particle:Td:耶)' # delim are ignored

        List format
            Used for :meth:`from_list`, and :meth:`to_list`.

            .. code-block:: python

                [
                    'S(Head:Nab:中文字|particle:Td:耶)', # parse clause
                    '，',                               # punctuations
                ]

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.

            .. code-block:: python

                {
                    'clause': 'S(Head:Nab:中文字|particle:Td:耶)', # parse clause
                    'delim': '，',                                # punctuations
                }
    """

    from_text = NotImplemented

    def to_text(self):
        return self.clause

    def to_tree(self):
        """Transform to tree format.

        Returns
        -------
            :class:`~.util.parse_tree.ParseTree`
                the tree format of this clause. (`None` if **clause** is `None`)

        .. seealso::
            :meth:`ParseTree.from_text() <.util.parse_tree.ParseTree.from_text>`.

        """
        return _ParseTree.from_text(self.clause) if self.clause else None

################################################################################################################################

class ParseSentence(_BaseList):
    """A parse sentence.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`to_text`.

            .. code-block:: python

                [ # delim are ignored
                    'S(Head:Nab:中文字|particle:Td:耶)',                    # Clause 1
                    '%(particle:I:啊|manner:Dh:哈|manner:Dh:哈|time:Dh:哈), # Clause 2
                ]

        List format
            Used for :meth:`from_list`, and :meth:`to_list`.

            .. code-block:: python

                [
                    [ # Clause 1
                        'S(Head:Nab:中文字|particle:Td:耶)',
                        '，',
                    ],
                    [ # Clause 2
                        '%(particle:I:啊|manner:Dh:哈|manner:Dh:哈|time:Dh:哈),
                        '。',
                    ],
                ]

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.

            .. code-block:: python

                [
                    { # Clause 1
                        'clause': 'S(Head:Nab:中文字|particle:Td:耶)',
                        'delim': '，',
                    },
                    { # Clause 2
                        'clause': '%(particle:I:啊|manner:Dh:哈|manner:Dh:哈|time:Dh:哈),
                        'delim': '。',
                    },
                ]

    """

    from_text = NotImplemented

    item_class = ParseClause

class ParseParagraph(_BaseList):
    """A list of parse sentence.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`to_text`.

            .. code-block:: python

                [ # delim are ignored
                    [ # Sentence 1
                        'S(Head:Nab:中文字|particle:Td:耶)',
                        '%(particle:I:啊|manner:Dh:哈|manner:Dh:哈|time:Dh:哈),
                    ],
                    [ # Sentence 2
                        None,
                        'VP(Head:VH11:完蛋|particle:Ta:了),
                        'S(agent:NP(apposition:Nba:畢卡索|Head:Nhaa:他)|Head:VE2:想)',
                    ],
                ]

        List format
            Used for :meth:`from_list`, and :meth:`to_list`.

            .. code-block:: python

                [
                    [ # Sentence 1
                        [
                            'S(Head:Nab:中文字|particle:Td:耶)',
                            '，',
                        ],
                        [
                            '%(particle:I:啊|manner:Dh:哈|manner:Dh:哈|time:Dh:哈),
                            '。',
                        ],
                    ],
                    [ # Sentence 2
                        [
                            None,
                            '「',
                        ],
                        [
                            'VP(Head:VH11:完蛋|particle:Ta:了),
                            '！」',
                        ],
                        [
                            'S(agent:NP(apposition:Nba:畢卡索|Head:Nhaa:他)|Head:VE2:想)',
                            '',
                        ],
                    ],
                ]

        Dict format
            Used for :meth:`from_dict`, and :meth:`to_dict`.

            .. code-block:: python

                [
                    [ # Sentence 1
                        {
                            'clause': 'S(Head:Nab:中文字|particle:Td:耶)',
                            'delim': '，',
                        },
                        {
                            'clause': '%(particle:I:啊|manner:Dh:哈|manner:Dh:哈|time:Dh:哈),
                            'delim': '。',
                        },
                    ],
                    [ # Sentence 2
                        {
                            'clause': None,
                            'delim': '「',
                        },
                        {
                            'clause': 'VP(Head:VH11:完蛋|particle:Ta:了),
                            'delim': '！」',
                        },
                        {
                            'clause': 'S(agent:NP(apposition:Nba:畢卡索|Head:Nhaa:他)|Head:VE2:想)',
                            'delim': '',
                        },
                    ],
                ]
    """

    from_text = NotImplemented

    item_class = ParseSentence



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
The Official CKIP CoreNLP Toolkit.
"""

__author_name__ = 'Mu Yang'
__author_email__ = 'emfomy@gmail.com'
__copyright__ = '2018-2020 CKIP Lab'

__title__ = 'CKIPNLP'
__version__ = '1.0.0'
__description__ = 'CKIP CoreNLP'
__license__ = 'GPL-3.0'

__url__ = 'https://ckipnlp.readthedocs.io'
__download_url__ = __url__+'/tarball/'+__version__



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides base containers.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

import json as _json

from abc import (
    ABCMeta as _ABCMeta,
    abstractmethod as _abstractmethod,
)

from collections import (
    UserList as _UserList,
)

################################################################################################################################

class Base(metaclass=_ABCMeta):
    """The base CKIPNLP container."""

    ########################################################################################################################

    @classmethod
    @_abstractmethod
    def from_text(cls, data):
        """Construct an instance from text format.

        Parameters
        ----------
            data : str
        """
        return NotImplemented  # pragma: no cover

    @_abstractmethod
    def to_text(self):
        """Transform to plain text.

        Returns
        -------
            str
        """
        return NotImplemented  # pragma: no cover

    ########################################################################################################################

    @classmethod
    @_abstractmethod
    def from_list(cls, data):
        """Construct an instance from python built-in containers."""
        return NotImplemented  # pragma: no cover

    @_abstractmethod
    def to_list(self):
        """Transform to python built-in containers."""
        return NotImplemented  # pragma: no cover

    ########################################################################################################################

    @classmethod
    @_abstractmethod
    def from_dict(cls, data):
        """Construct an instance from python built-in containers."""
        return NotImplemented  # pragma: no cover

    @_abstractmethod
    def to_dict(self):
        """Transform to python built-in containers."""
        return NotImplemented  # pragma: no cover

    ########################################################################################################################

    @classmethod
    def from_json(cls, data, **kwargs):
        """Construct an instance from JSON format.

        Parameters
        ----------
            data : str
                please refer :meth:`from_dict` for format details.
        """
        return cls.from_dict(_json.loads(data, **kwargs))

    def to_json(self, **kwargs):
        """Transform to JSON format.

        Returns
        -------
            str
        """
        return _json.dumps(self.to_dict(), **kwargs)

################################################################################################################################

class BaseTuple(Base, metaclass=_ABCMeta):
    """The base CKIPNLP tuple."""

    ########################################################################################################################

    @classmethod
    @_abstractmethod
    def from_text(cls, data):
        return NotImplemented  # pragma: no cover

    @_abstractmethod
    def to_text(self):
        return NotImplemented  # pragma: no cover

    ########################################################################################################################

    @classmethod
    def from_list(cls, data):
        """Construct an instance from python built-in containers.

        Parameters
        ----------
            data : list
        """
        return cls(*data)

    def to_list(self):
        """Transform to python built-in containers.

        Returns
        -------
            list
        """
        return list(self)

    ########################################################################################################################

    @classmethod
    def from_dict(cls, data):
        """Construct an instance from python built-in containers.

        Parameters
        ----------
            data : dict
        """
        return cls(**data)

    def to_dict(self):
        """Transform to python built-in containers.

        Returns
        -------
            dict
        """
        return dict(self._asdict())  # pylint: disable=no-member

################################################################################################################################

class _BaseList(Base, _UserList):
    """The base CKIPNLP list."""

    item_class = NotImplemented

    ########################################################################################################################

    @classmethod
    def from_text(cls, data):
        """Construct an instance from text format.

        Parameters
        ----------
            data : Sequence[str]
                list of texts as ``item_class.from_text`` input.
        """
        return cls(map(cls._item_from_text, data))  # pylint: disable=no-member

    def to_text(self):
        """Transform to plain text.

        Returns
        -------
            List[str]
        """
        return list(map(self._item_to_text, self))  # pylint: disable=no-member

    ########################################################################################################################

    @classmethod
    def from_list(cls, data):
        """Construct an instance from python built-in containers.

        Parameters
        ----------
            data : Sequence[Container]
                list of objects as ``item_class.from_list`` input.
        """
        return cls(map(cls._item_from_list, data))  # pylint: disable=no-member

    def to_list(self):
        """Transform to python built-in containers.

        Returns
        -------
            List[Container]
        """
        return list(map(self._item_to_list, self))  # pylint: disable=no-member

    ########################################################################################################################

    @classmethod
    def from_dict(cls, data):
        """Construct an instance from python built-in containers.

        Parameters
        ----------
            data : Sequence[Container]
                list of objects as ``item_class.from_dict`` input.
        """
        return cls(map(cls._item_from_dict, data))  # pylint: disable=no-member

    def to_dict(self):
        """Transform to python built-in containers.

        Returns
        -------
            List[Container]
        """
        return list(map(self._item_to_dict, self))  # pylint: disable=no-member

################################################################################################################################

class _BaseSentence(_BaseList):
    """The base CKIPNLP sentence."""

    @classmethod
    def from_text(cls, data):
        """Construct an instance from text format.

        Parameters
        ----------
            data : str
                words segmented by ``'\\u3000'``.
        """
        return cls(map(cls._item_from_text, data.split('\u3000')))  # pylint: disable=no-member

    def to_text(self):
        return '\u3000'.join(map(self._item_to_text, self))  # pylint: disable=no-member

################################################################################################################################

class _InterfaceItem:
    """Container has ckipnlp item class."""

    @classmethod
    def _item_from_text(cls, data):
        return cls.item_class.from_text(data)  # pylint: disable=no-member

    @classmethod
    def _item_to_text(cls, item):
        return cls.item_class.to_text(item)  # pylint: disable=no-member

    @classmethod
    def _item_from_list(cls, data):
        return cls.item_class.from_list(data)  # pylint: disable=no-member

    @classmethod
    def _item_to_list(cls, item):
        return cls.item_class.to_list(item)  # pylint: disable=no-member

    @classmethod
    def _item_from_dict(cls, data):
        return cls.item_class.from_dict(data)  # pylint: disable=no-member

    @classmethod
    def _item_to_dict(cls, item):
        return cls.item_class.to_dict(item)  # pylint: disable=no-member

################################################################################################################################

class _InterfaceBuiltInItem:
    @classmethod
    def _item_from_text(cls, data):
        return cls.item_class(data)  # pylint: disable=no-member

    @classmethod
    def _item_to_text(cls, item):
        return cls.item_class(item)  # pylint: disable=no-member

    @classmethod
    def _item_from_list(cls, data):
        return cls.item_class(data)  # pylint: disable=no-member

    @classmethod
    def _item_to_list(cls, item):
        return cls.item_class(item)  # pylint: disable=no-member

    @classmethod
    def _item_from_dict(cls, data):
        return cls.item_class(data)  # pylint: disable=no-member

    @classmethod
    def _item_to_dict(cls, item):
        return cls.item_class(item)  # pylint: disable=no-member

################################################################################################################################

class BaseList(_BaseList, _InterfaceItem):
    """The base CKIPNLP list.

    .. attribute:: item_class
        :value: Not Implemented

        Must be a CKIPNLP container class.
    """

class BaseList0(_BaseList, _InterfaceBuiltInItem):
    """The base CKIPNLP list with built-in item class.

    .. attribute:: item_class
        :value: Not Implemented

        Must be a built-in type.
    """

class BaseSentence(_BaseSentence, _InterfaceItem):
    """The base CKIPNLP sentence.

    .. attribute:: item_class
        :value: Not Implemented

        Must be a CKIPNLP container class.
    """

class BaseSentence0(_BaseSentence, _InterfaceBuiltInItem):
    """The base CKIPNLP sentence with built-in item class.

    .. attribute:: item_class
        :value: Not Implemented

        Must be a built-in type.
    """



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides coreference resolution pipeline.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from collections.abc import (
    Mapping as _Mapping,
)

from ckipnlp.driver.base import (
    DriverRegister as _DriverRegister,
)

from .kernel import (
    CkipPipeline as _CkipPipeline,
)

################################################################################################################################

class CkipCorefDocument(_Mapping):
    """The coreference document.

    Attributes
    ----------
        ws : :class:`~ckipnlp.container.seg.SegParagraph`
            The word-segmented sentences.
        pos : :class:`~ckipnlp.container.seg.SegParagraph`
            The part-of-speech sentences.
        conparse : :class:`~ckipnlp.container.parse.ParseParagraph`
            The constituency sentences.
        coref : :class:`~ckipnlp.container.coref.CorefParagraph`
            The coreference resolution results.
    """

    __keys = ('ws', 'pos', 'conparse', 'coref',)

    def __init__(self, *, ws=None, pos=None, conparse=None, coref=None):
        self.ws = ws
        self.pos = pos
        self.conparse = conparse
        self.coref = coref

    def __len__(self):
        return len(self.__keys)

    def __iter__(self):
        yield from self.__keys

    def __getitem__(self, key):
        return getattr(self, key)

################################################################################################################################

class CkipCorefPipeline(_CkipPipeline):
    """The coreference resolution pipeline.

    Arguments
    ---------
        sentence_segmenter : str
            The type of sentence segmenter.

        word_segmenter : str
            The type of word segmenter.

        pos_tagger : str
            The type of part-of-speech tagger.

        ner_chunker : str
            The type of named-entity recognition chunker.

        con_parser : str
            The type of constituency parser.

        coref_chunker : str
            The type of coreference resolution chunker.

    Other Parameters
    ----------------
        lazy : bool
            Lazy initialize the drivers.

        opts : Dict[str, Dict]
            The driver options. Key: driver name (e.g. `'sentence_segmenter'`); Value: a dictionary of options.
    """

    def __init__(self, *,
        coref_chunker='default',
        lazy=True,
        opts={},
        **kwargs,
    ):
        super().__init__(lazy=lazy, opts=opts, **kwargs)

        # CoRef
        if coref_chunker:
            assert self._wspos_driver.is_dummy, 'Coreference pipeline is not compatible with CkipClassic word segmenter!'

        self._coref_chunker = _DriverRegister.get('coref_chunker', coref_chunker)(
            lazy=lazy, **opts.get('coref_chunker', {}),
        )

    def __call__(self, doc):
        """Apply coreference delectation.

        Arguments
        ---------
            doc : :class:`~.kernel.CkipDocument`
                The input document.

        Returns
        -------
            corefdoc : :class:`CkipCorefDocument`
                The coreference document.

        .. note::

            **doc** is also modified if necessary dependencies (**ws**, **pos**, **ner**) is not computed yet.
        """

        corefdoc = CkipCorefDocument()
        self.get_coref(doc, corefdoc)
        return corefdoc

    def get_coref(self, doc, corefdoc):
        """Apply coreference delectation.

        Arguments
        ---------
            doc : :class:`~.kernel.CkipDocument`
                The input document.
            corefdoc : :class:`CkipCorefDocument`
                The input document for coreference.

        Returns
        -------
            corefdoc.coref : :class:`~ckipnlp.container.coref.CorefParagraph`
                The coreference results.

        .. note::

            This routine modify **corefdoc** inplace.

            **doc** is also modified if necessary dependencies (**ws**, **pos**, **ner**) is not computed yet.
        """
        self.get_text(doc)
        self.get_ws(doc)
        self.get_pos(doc)
        self.get_ner(doc)

        # Update word segmentation
        if corefdoc.ws is None:
            corefdoc.ws = self._coref_chunker.transform_ws(
                text=doc.text,
                ws=doc.ws,
                ner=doc.ner,
            )

        # Update POS-tagging
        if corefdoc.pos is None:
            corefdoc.pos = self.get_pos(corefdoc)
            self._coref_chunker.transform_pos(
                ws=corefdoc.ws,
                pos=corefdoc.pos,
                ner=doc.ner,
            )

        # Do parsing
        if corefdoc.conparse is None:
            corefdoc.conparse = self.get_conparse(corefdoc)

        # Do coreference resolution
        corefdoc.coref = self._coref_chunker(conparse=corefdoc.conparse)

        return corefdoc.coref



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides containers for text sentences.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from .base import (
    BaseList0 as _BaseList0
)

################################################################################################################################

class TextParagraph(_BaseList0):
    """A list of text sentence.

    .. admonition:: Data Structure Examples

        Text/List/Dict format
            Used for :meth:`from_text`, :meth:`to_text`, :meth:`from_list`, :meth:`to_list`, :meth:`from_dict`, and
            :meth:`to_dict`.

            .. code-block:: python

                [
                    '中文字耶，啊哈哈哈。',    # Sentence 1
                    '「完蛋了！」畢卡索他想', # Sentence 2
                ]
    """

    item_class = str



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module implements specialized container datatypes for CKIPNLP.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from .text import *
from .seg import *
from .ner import *
from .parse import *
from .coref import *

from .util.wspos import *
from .util.parse_tree import *



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides drivers with CkipClassic backend.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from abc import (
    abstractmethod as _abstractmethod,
)

from itertools import (
    chain as _chain,
)

from ckipnlp.container import (
    TextParagraph as _TextParagraph,
    SegParagraph as _SegParagraph,
    WsPosSentence as _WsPosSentence,
    WsPosParagraph as _WsPosParagraph,
    ParseParagraph as _ParseParagraph,
)

from .base import (
    BaseDriver as _BaseDriver,
)

################################################################################################################################

class CkipClassicWordSegmenter(_BaseDriver):
    """The CKIP word segmentation driver with CkipClassic backend.

    Arguments
    ---------
        lazy : bool
            Lazy initialize the driver.
        do_pos : bool
            Returns POS-tag or not
        lexicons: Iterable[Tuple[str, str]]
            A list of the lexicon words and their POS-tags.

    .. method:: __call__(*, text)

        Apply word segmentation.

        Parameters
            **text** (:class:`TextParagraph <ckipnlp.container.text.TextParagraph>`) — The sentences.

        Returns
            - **ws** (:class:`TextParagraph <ckipnlp.container.text.TextParagraph>`) — The word-segmented sentences.
            - **pos** (:class:`TextParagraph <ckipnlp.container.text.TextParagraph>`) — The part-of-speech sentences.
              (returns if **do_pos** is set.)
    """

    driver_type = None
    driver_family = 'classic'
    driver_inputs = None

    _count = 0

    def __init__(self, *, lazy=False, do_pos=False, lexicons=None):
        super().__init__(lazy=lazy)
        self._do_pos = do_pos
        self._lexicons = lexicons

    def _init(self):
        self.__class__._count += 1  # pylint: disable=protected-access
        if self.__class__._count > 1:  # pylint: disable=protected-access
            raise RuntimeError(f'Never instance more than one {self.__class__.__name__}!')

        import ckip_classic.ws
        self._core = ckip_classic.ws.CkipWs(
            new_style_format=True,
            lex_list=self._lexicons,
        )

    def _call(self, *, text):
        assert isinstance(text, _TextParagraph)

        wspos_text = self._core.apply_list(text.to_text())
        ws, pos = _WsPosParagraph.from_text(wspos_text)

        return (ws, pos,) if self._do_pos else ws

class _CkipClassicWordSegmenter(CkipClassicWordSegmenter):
    """The dummy class for :class:`CkipClassicWordSegmenter` for pipeline."""

    driver_type = 'word_segmenter'
    driver_family = 'classic'
    driver_inputs = ('text',)

    def __init__(self, *, lazy=False, lexicons=None):
        super().__init__(lazy=lazy, do_pos=False, lexicons=lexicons)

class _CkipClassic2WsPos(CkipClassicWordSegmenter):
    """The dummy class for :class:`CkipClassicWordSegmenter` for pipeline."""

    driver_type = '_wspos'
    driver_family = '_classic'
    driver_inputs = ('text',)

    def __init__(self, *, lazy=False, lexicons=None):
        super().__init__(lazy=lazy, do_pos=True, lexicons=lexicons)

class _CkipClassic2WordSegmenter(_BaseDriver):
    """The dummy class for :class:`CkipClassicWordSegmenter` for pipeline."""

    driver_type = 'word_segmenter'
    driver_family = '_classic'
    driver_inputs = ('_wspos',)

    def _init(self):
        pass

    def _call(self, *, _wspos):
        return _wspos[0]

class _CkipClassic2PosTagger(_BaseDriver):
    """The dummy class for :class:`CkipClassicWordSegmenter` for pipeline."""

    driver_type = 'pos_tagger'
    driver_family = '_classic'
    driver_inputs = ('_wspos',)

    def _init(self):
        pass

    def _call(self, *, _wspos):
        return _wspos[1]

################################################################################################################################

class _CkipClassicConParser(_BaseDriver):


    driver_type = 'con_parser'
    driver_inputs = ('ws', 'pos',)

    @_abstractmethod
    def driver_family(self):  # pylint: disable=missing-docstring
        return NotImplemented

    @_abstractmethod
    def _init(self):
        return NotImplemented

    def _call(self, *, ws, pos):
        assert isinstance(ws, _SegParagraph)
        assert isinstance(pos, _SegParagraph)


        conparse_text = []
        for ws_sent, pos_sent in zip(ws, pos):
            conparse_sent_text = []
            ws_clause = []
            pos_clause = []
            for ws_token, pos_token in _chain(zip(ws_sent, pos_sent), [(None, None),]):

                # Skip WHITESPACE
                if pos_token == 'WHITESPACE':
                    continue

                # Segment clauses by punctuations
                if pos_token is None or (pos_token.endswith('CATEGORY') and pos_token != 'PAUSECATEGORY'):
                    if ws_clause:
                        wspos_clause_text = _WsPosSentence.to_text(ws_clause, pos_clause)
                        for conparse_clause_text in self._core.apply_list([wspos_clause_text]):
                            conparse_sent_text.append([self._normalize(conparse_clause_text), '',])

                    if ws_token:
                        if not conparse_sent_text:
                            conparse_sent_text.append([None, '',])
                        conparse_sent_text[-1][1] += ws_token

                    ws_clause = []
                    pos_clause = []

                else:
                    ws_clause.append(self._half2full(ws_token))
                    pos_clause.append(pos_token)

            conparse_text.append(conparse_sent_text)
        conparse = _ParseParagraph.from_list(conparse_text)

        return conparse

    @staticmethod
    def _half2full(text):
        return text \
           .replace('(', '（') \
           .replace(')', '）') \
           .replace('+', '＋') \
           .replace('-', '－') \
           .replace(':', '：') \
           .replace('|', '｜')

    @staticmethod
    def _normalize(text):
        return text.split('] ', 2)[-1].rstrip('#')

################################################################################################################################

class CkipClassicConParser(_CkipClassicConParser):
    """The CKIP constituency parsing driver with CkipClassic backend.

    Arguments
    ---------
        lazy : bool
            Lazy initialize the driver.

    .. method:: __call__(*, ws, pos)

        Apply constituency parsing.

        Parameters
            - **ws** (:class:`~ckipnlp.container.text.TextParagraph`) — The word-segmented sentences.
            - **pos** (:class:`~ckipnlp.container.text.TextParagraph`) — The part-of-speech sentences.

        Returns
            **conparse** (:class:`~ckipnlp.container.parse.ParseSentence`) — The constituency-parsing sentences.
    """

    driver_family = 'classic'

    _count = 0

    def _init(self):

        self.__class__._count += 1  # pylint: disable=protected-access
        if self.__class__._count > 1:  # pylint: disable=protected-access
            raise RuntimeError(f'Never instance more than one {self.__class__.__name__}!')

        import ckip_classic.parser
        self._core = ckip_classic.parser.CkipParser(do_ws=False)

class CkipClassicConParserClient(_CkipClassicConParser):
    """The CKIP constituency parsing driver with CkipClassic client backend.

    Arguments
    ---------
        lazy : bool
            Lazy initialize the driver.

    Notes
    -----

        Please register an account at http://parser.iis.sinica.edu.tw/v1/reg.exe and
        set the environment variables ``$CKIPPARSER_USERNAME`` and ``$CKIPPARSER_PASSWORD``.

    .. method:: __call__(*, ws, pos)

        Apply constituency parsing.

        Parameters
            - **ws** (:class:`~ckipnlp.container.text.TextParagraph`) — The word-segmented sentences.
            - **pos** (:class:`~ckipnlp.container.text.TextParagraph`) — The part-of-speech sentences.

        Returns
            **conparse** (:class:`~ckipnlp.container.parse.ParseSentence`) — The constituency-parsing sentences.

    """

    driver_family = 'classic-client'

    def _init(self):

        import ckip_classic.client
        self._core = ckip_classic.client.CkipParserClient()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from setuptools import setup, find_namespace_packages
import ckipnlp as about

################################################################################

def main():

    with open('README.rst', encoding='utf-8') as fin:
        readme = fin.read()

    setup(
        name='ckipnlp',
        version=about.__version__,
        author=about.__author_name__,
        author_email=about.__author_email__,
        description=about.__description__,
        long_description=readme,
        long_description_content_type='text/x-rst',
        url=about.__url__,
        download_url=about.__download_url__,
        platforms=['linux_x86_64'],
        license=about.__license__,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3 :: Only',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Operating System :: POSIX :: Linux',
            'Natural Language :: Chinese (Traditional)',
        ],
        python_requires='>=3.6',
        packages=find_namespace_packages(include=['ckipnlp', 'ckipnlp.*',]),
        install_requires=[
            'appdirs>=1.4.3',
            'numpy>=1.18',
            'treelib>=1.5.5',
        ],
        extras_require={
            'classic': ['ckip-classic>=1.1.2'],
            'tagger': ['ckiptagger[tf]>=0.2.1'],
            'tagger-gpu': ['ckiptagger[tfgpu]>=0.2.1'],
        },
        data_files=[],
    )

################################################################################

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides containers for coreference sentences.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from typing import (
    NamedTuple as _NamedTuple,
    Tuple as _Tuple,
)

from .base import (
    BaseTuple as _BaseTuple,
    BaseList as _BaseList,
    BaseSentence as _BaseSentence,
)

################################################################################################################################

class _CorefToken(_NamedTuple):
    word: str
    coref: _Tuple[int, str,]
    idx: int

class CorefToken(_BaseTuple, _CorefToken):
    """A coreference token.

    Attributes
    ----------
        word : str
            the token word.
        coref : Tuple[int, str]
            the coreference ID and type. `None` if not a coreference source or target.

            * **type**:
                * `'source'`: coreference source.
                * `'target'`: coreference target.
                * `'zero'`: null element coreference target.

        idx : Tuple[int, int]
            the node indexes (clause index, token index) in parse tree.
            **idx[1]** = `None` if this node is a null element or the punctuations.

    Note
    ----
        This class is an subclass of :class:`tuple`. To change the attribute, please create a new instance instead.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`to_text`.

            .. code-block:: python

                '畢卡索_0'

        List format
            Used for :meth:`from_list` and :meth:`to_list`.

            .. code-block:: python

                [
                    '畢卡索',       # token word
                    (0, 'source'), # coref ID and type
                    (2, 2),        # node index
                ]

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.

            .. code-block:: python

                {
                    'word': '畢卡索',        # token word
                    'coref': (0, 'source'), # coref ID and type
                    'idx': (2, 2),          # node index
                }
    """

    def __new__(cls, word, coref, idx, **kwargs):  # pylint: disable=signature-differs
        return super().__new__(cls, word, tuple(coref) if coref else None, tuple(idx), **kwargs)

    from_text = NotImplemented

    def to_text(self):
        return f'{self.word}_{self.coref[0]}' if self.coref else self.word

################################################################################################################################

class CorefSentence(_BaseSentence):
    """A list of coreference sentence.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`to_text`.

            .. code-block:: python

                # Token segmented by \\u3000 (full-width space)
                '「\u3000完蛋\u3000了\u3000！」\u3000，\u3000畢卡索_0\u3000他_0\u3000想'

        List format
            Used for :meth:`from_list` and :meth:`to_list`.

            .. code-block:: python

                [
                    [ '「', None, (0, None,), ],
                    [ '完蛋', None, (1, 0,), ],
                    [ '了', None, (1, 1,), ],
                    [ '！」', None, (1, None,), ],
                    [ '畢卡索', (0, 'source'), (2, 2,), ],
                    [ '他', (0, 'target'), (2, 3,), ],
                    [ '想', None, (2, 4,), ],
                ]

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.

            .. code-block:: python

                [
                    { 'word': '「', 'coref': None, 'idx': (0, None,), },
                    { 'word': '完蛋', 'coref': None, 'idx': (1, 0,), },
                    { 'word': '了', 'coref': None, 'idx': (1, 1,), },
                    { 'word': '！」', 'coref': None, 'idx': (1, None,), },
                    { 'word': '畢卡索', 'coref': (0, 'source'), 'idx': (2, 2,), },
                    { 'word': '他', 'coref': (0, 'target'), 'idx': (2, 3,), },
                    { 'word': '想', 'coref': None, 'idx': (2, 4,), },
                ]
    """

    item_class = CorefToken

    from_text = NotImplemented

################################################################################################################################

class CorefParagraph(_BaseList):
    """A list of coreference sentence.

    .. admonition:: Data Structure Examples

        Text format
            Used for :meth:`to_text`.

            .. code-block:: python

                [
                    '「\u3000完蛋\u3000了\u3000！」\u3000，\u3000畢卡索_0\u3000他_0\u3000想', # Sentence 1
                    '但是\u3000None_0\u3000也\u3000沒有\u3000辦法', # Sentence 1
                ]

        List format
            Used for :meth:`from_list` and :meth:`to_list`.

            .. code-block:: python

                [
                    [ # Sentence 1
                        [ '「', None, (0, None,), ],
                        [ '完蛋', None, (1, 0,), ],
                        [ '了', None, (1, 1,), ],
                        [ '！」', None, (1, None,), ],
                        [ '畢卡索', (0, 'source'), (2, 2,), ],
                        [ '他', (0, 'target'), (2, 3,), ],
                        [ '想', None, (2, 4,), ],
                    ],
                    [ # Sentence 2
                        [ '但是', None, (0, 1,), ],
                        [ None, (0, 'zero'), (0, None,), ],
                        [ '也', None, (0, 2,), ],
                        [ '沒有', None, (0, 3,), ],
                        [ '辦法', None, (0, 5,), ],
                    ],
                ]

        Dict format
            Used for :meth:`from_dict` and :meth:`to_dict`.

            .. code-block:: python

                [
                    [ # Sentence 1
                        { 'word': '「', 'coref': None, 'idx': (0, None,), },
                        { 'word': '完蛋', 'coref': None, 'idx': (1, 0,), },
                        { 'word': '了', 'coref': None, 'idx': (1, 1,), },
                        { 'word': '！」', 'coref': None, 'idx': (1, None,), },
                        { 'word': '畢卡索', 'coref': (0, 'source'), 'idx': (2, 2,), },
                        { 'word': '他', 'coref': (0, 'target'), 'idx': (2, 3,), },
                        { 'word': '想', 'coref': None, 'idx': (2, 4,), },
                    ],
                    [ # Sentence 2
                        { 'word': '但是', 'coref': None, 'idx': (0, 1,), },
                        { 'word': None, 'coref': (0, 'zero'), 'idx': (0, None,), },
                        { 'word': '也', 'coref': None, 'idx': (0, 2,), },
                        { 'word': '沒有', 'coref': None, 'idx': (0, 3,), },
                        { 'word': '辦法', 'coref': None, 'idx': (0, 5,), },
                    ],
                ]
    """

    item_class = CorefSentence

    from_text = NotImplemented



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides drivers with CkipTagger backend.
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from ckipnlp.container import (
    TextParagraph as _TextParagraph,
    SegParagraph as _SegParagraph,
    NerParagraph as _NerParagraph,
)

from ckipnlp.util.data import (
    get_tagger_data as _get_tagger_data,
)

from .base import (
    BaseDriver as _BaseDriver,
)

################################################################################################################################

class CkipTaggerWordSegmenter(_BaseDriver):
    """The CKIP word segmentation driver with CkipTagger backend.

    Arguments
    ---------
        lazy : bool
            Lazy initialize the driver.
        disable_cuda : bool
            Disable GPU usage.
        recommend_lexicons: Mapping[str, float]
            A mapping of lexicon words to their relative weights.
        coerce_lexicons: Mapping[str, float]
            A mapping of lexicon words to their relative weights.

    Other Parameters
    ----------------
        **opts
            Extra options for :meth:`ckiptagger.WS.__call__`.
            (Please refer https://github.com/ckiplab/ckiptagger#4-run-the-ws-pos-ner-pipeline for details.)


    .. method:: __call__(*, text)

        Apply word segmentation.

        Parameters
            **text** (:class:`TextParagraph <ckipnlp.container.text.TextParagraph>`) — The sentences.

        Returns
            **ws** (:class:`TextParagraph <ckipnlp.container.text.TextParagraph>`) — The word-segmented sentences.
    """

    driver_type = 'word_segmenter'
    driver_family = 'tagger'
    driver_inputs = ('text',)

    def __init__(self, *, lazy=False, disable_cuda=True, recommend_lexicons={}, coerce_lexicons={}, **opts):
        super().__init__(lazy=lazy)
        self._disable_cuda = disable_cuda
        self._recommend_lexicons = recommend_lexicons
        self._coerce_lexicons = coerce_lexicons
        self._opts = opts

    def _init(self):
        import ckiptagger

        if self._recommend_lexicons:
            self._opts['recommend_dictionary'] = ckiptagger.construct_dictionary(self._recommend_lexicons)
        if self._coerce_lexicons:
            self._opts['coerce_dictionary'] = ckiptagger.construct_dictionary(self._coerce_lexicons)

        self._core = ckiptagger.WS(_get_tagger_data(), disable_cuda=self._disable_cuda)

    def _call(self, *, text):
        assert isinstance(text, _TextParagraph)

        ws_list = self._core(text, **self._opts)
        ws = _SegParagraph.from_list(ws_list)

        return ws

################################################################################################################################

class CkipTaggerPosTagger(_BaseDriver):
    """The CKIP part-of-speech tagging driver with CkipTagger backend.

    Arguments
    ---------
        lazy : bool
            Lazy initialize the driver.
        disable_cuda : bool
            Disable GPU usage.

    Other Parameters
    ----------------
        **opts
            Extra options for :meth:`ckiptagger.POS.__call__`.
            (Please refer https://github.com/ckiplab/ckiptagger#4-run-the-ws-pos-ner-pipeline for details.)

    .. method:: __call__(*, text)

        Apply part-of-speech tagging.

        Parameters
            **ws** (:class:`TextParagraph <ckipnlp.container.text.TextParagraph>`) — The word-segmented sentences.

        Returns
            **pos** (:class:`TextParagraph <ckipnlp.container.text.TextParagraph>`) — The part-of-speech sentences.
    """

    driver_type = 'pos_tagger'
    driver_family = 'tagger'
    driver_inputs = ('ws',)

    def __init__(self, *, lazy=False, disable_cuda=True, **opts):
        super().__init__(lazy=lazy)
        self._disable_cuda = disable_cuda
        self._opts = opts

    def _init(self):
        import ckiptagger
        self._core = ckiptagger.POS(_get_tagger_data(), disable_cuda=self._disable_cuda)

    def _call(self, *, ws):
        assert isinstance(ws, _SegParagraph)

        pos_list = self._core(ws, **self._opts)
        pos = _SegParagraph.from_list(pos_list)

        return pos

################################################################################################################################

class CkipTaggerNerChunker(_BaseDriver):
    """The CKIP named-entity recognition driver with CkipTagger backend.

    Arguments
    ---------
        lazy : bool
            Lazy initialize the driver.
        disable_cuda : bool
            Disable GPU usage.

    Other Parameters
    ----------------
        **opts
            Extra options for :meth:`ckiptagger.NER.__call__`.
            (Please refer https://github.com/ckiplab/ckiptagger#4-run-the-ws-pos-ner-pipeline for details.)

    .. method:: __call__(*, text)

        Apply named-entity recognition.

        Parameters
            - **ws** (:class:`TextParagraph <ckipnlp.container.text.TextParagraph>`) — The word-segmented sentences.
            - **pos** (:class:`TextParagraph <ckipnlp.container.text.TextParagraph>`) — The part-of-speech sentences.

        Returns
            **ner** (:class:`NerParagraph <ckipnlp.container.ner.NerParagraph>`) — The named-entity recognition results.
    """

    driver_type = 'ner_tagger'
    driver_family = 'tagger'
    driver_inputs = ('ws', 'pos',)

    def __init__(self, *, lazy=False, disable_cuda=True, **opts):
        super().__init__(lazy=lazy)
        self._disable_cuda = disable_cuda
        self._opts = opts

    def _init(self):
        import ckiptagger
        self._core = ckiptagger.NER(_get_tagger_data(), disable_cuda=self._disable_cuda)

    def _call(self, *, ws, pos):
        assert isinstance(ws, _SegParagraph)
        assert isinstance(pos, _SegParagraph)

        ner_list = self._core(ws, pos, **self._opts)
        ner = _NerParagraph.from_tagger(ner_list)

        return ner
